@@ -10,7 +10,8 @@ from aiohttp import hdrs
 from aiohttp import http
 from aiohttp.client import ClientSession
 from aiohttp.client_reqrep import ClientResponse
-from ddt import ddt, data
+from ddt import ddt, data, unpack
+from packaging.version import Version
 
 try:
     from aiohttp.errors import (
@@ -25,16 +26,13 @@ except ImportError:
     )
     from aiohttp.http_exceptions import HttpProcessingError
 
-from aioresponses.compat import (
-    AIOHTTP_VERSION, URL,
-    fail_on,
-    skipIf,
-    AsyncTestCase
-)
-
+from aioresponses.compat import AIOHTTP_VERSION, URL
 from aioresponses import CallbackResult, aioresponses
+from .base import fail_on, skipIf, AsyncTestCase
 
 
+MINIMUM_AIOHTTP_VERSION = Version('3.4.0')
+
 @ddt
 class AIOResponsesTestCase(AsyncTestCase):
 
@@ -82,6 +80,33 @@ class AIOResponsesTestCase(AsyncTestCase):
         self.assertIsInstance(response, ClientResponse)
         self.assertEqual(response.status, 204)
 
+    @unpack
+    @data(
+        ("http://example.com", "/api?foo=bar#fragment"),
+        ("http://example.com/", "/api?foo=bar#fragment")
+    )
+    @aioresponses()
+    @skipIf(condition=AIOHTTP_VERSION < Version('3.8.0'), reason='aiohttp must be >= 3.8.0')
+    async def test_base_url(self, base_url, relative_url, m):
+        m.get(self.url, status=200)
+        self.session = ClientSession(base_url=base_url)
+        response = await self.session.get(relative_url)
+        self.assertEqual(response.status, 200)
+
+    @aioresponses()
+    @skipIf(condition=AIOHTTP_VERSION < Version('3.8.0'), reason='aiohttp must be >= 3.8.0')
+    async def test_session_headers(self, m):
+        m.get(self.url)
+        self.session = ClientSession(headers={"Authorization": "Bearer foobar"})
+        response = await self.session.get(self.url)
+
+        self.assertEqual(response.status, 200)
+
+        # Check that the headers from the ClientSession are within the request
+        key = ('GET', URL(self.url))
+        request = m.requests[key][0]
+        self.assertEqual(request.kwargs["headers"]["Authorization"], 'Bearer foobar')
+
     @aioresponses()
     async def test_returned_response_headers(self, m):
         m.get(self.url,
@@ -122,7 +147,7 @@ class AIOResponsesTestCase(AsyncTestCase):
         self.assertEqual(cm.exception.message, http.RESPONSES[400][0])
 
     @aioresponses()
-    @skipIf(condition=AIOHTTP_VERSION < '3.4.0',
+    @skipIf(condition=AIOHTTP_VERSION < MINIMUM_AIOHTTP_VERSION,
             reason='aiohttp<3.4.0 does not support raise_for_status '
                    'arguments for requests')
     async def test_request_raise_for_status(self, m):
@@ -147,6 +172,9 @@ class AIOResponsesTestCase(AsyncTestCase):
         )
         self.assertIsInstance(response, ClientResponse)
         self.assertEqual(response.status, 200)
+        self.assertEqual(len(m.requests), 2)
+        with self.assertRaises(AssertionError):
+            m.assert_called_once()
 
     @aioresponses()
     def test_method_dont_match(self, m):
@@ -154,6 +182,59 @@ class AIOResponsesTestCase(AsyncTestCase):
         with self.assertRaises(ClientConnectionError):
             self.run_async(self.session.post(self.url))
 
+    @aioresponses()
+    def test_post_with_data(self, m: aioresponses):
+        body = {'foo': 'bar'}
+        payload = {'spam': 'eggs'}
+        user_agent = {'User-Agent': 'aioresponses'}
+        m.post(
+            self.url,
+            payload=payload,
+            headers=dict(connection='keep-alive'),
+            body=body,
+        )
+        response = self.run_async(
+            self.session.post(
+                self.url,
+                data=payload,
+                headers=user_agent
+            )
+        )
+        self.assertIsInstance(response, ClientResponse)
+        self.assertEqual(response.status, 200)
+        response_data = self.run_async(response.json())
+        self.assertEqual(response_data, payload)
+        m.assert_called_once_with(
+            self.url,
+            method='POST',
+            data=payload,
+            headers={'User-Agent': 'aioresponses'}
+        )
+        # Wrong data
+        with self.assertRaises(AssertionError):
+            m.assert_called_once_with(
+                self.url,
+                method='POST',
+                data=body,
+                headers={'User-Agent': 'aioresponses'}
+            )
+        # Wrong url
+        with self.assertRaises(AssertionError):
+            m.assert_called_once_with(
+                'http://httpbin.org/',
+                method='POST',
+                data=payload,
+                headers={'User-Agent': 'aioresponses'}
+            )
+        # Wrong headers
+        with self.assertRaises(AssertionError):
+            m.assert_called_once_with(
+                self.url,
+                method='POST',
+                data=payload,
+                headers={'User-Agent': 'aiorequest'}
+            )
+
     @aioresponses()
     async def test_streaming(self, m):
         m.get(self.url, body='Test')
@@ -170,6 +251,26 @@ class AIOResponsesTestCase(AsyncTestCase):
         content = await resp.content.read(2)
         self.assertEqual(content, b'st')
 
+    @aioresponses()
+    async def test_binary_body(self, m):
+        body = b'Invalid utf-8: \x95\x00\x85'
+        m.get(self.url, body=body)
+        resp = await self.session.get(self.url)
+        content = await resp.read()
+        self.assertEqual(content, body)
+
+    @aioresponses()
+    async def test_binary_body_via_callback(self, m):
+        body = b'\x00\x01\x02\x80\x81\x82\x83\x84\x85'
+
+        def callback(url, **kwargs):
+            return CallbackResult(body=body)
+
+        m.get(self.url, callback=callback)
+        resp = await self.session.get(self.url)
+        content = await resp.read()
+        self.assertEqual(content, body)
+
     async def test_mocking_as_context_manager(self):
         with aioresponses() as aiomock:
             aiomock.add(self.url, payload={'foo': 'bar'})
@@ -255,6 +356,15 @@ class AIOResponsesTestCase(AsyncTestCase):
             with self.assertRaises(HttpProcessingError):
                 await self.session.get(url)
 
+            callback_called = asyncio.Event()
+            url = 'http://example.com/HttpProcessingError'
+            aiomock.get(url, exception=HttpProcessingError(message='foo'),
+                        callback=lambda *_, **__: callback_called.set())
+            with self.assertRaises(HttpProcessingError):
+                await self.session.get(url)
+
+            await callback_called.wait()
+
     async def test_multiple_requests(self):
         """Ensure that requests are saved the way they would have been sent."""
         with aioresponses() as m:
@@ -467,6 +577,77 @@ class AIOResponsesTestCase(AsyncTestCase):
         data = self.run_async(response.read())
         assert data == body
 
+    @aioresponses()
+    def test_assert_not_called(self, m: aioresponses):
+        m.get(self.url)
+        m.assert_not_called()
+        self.run_async(self.session.get(self.url))
+        with self.assertRaises(AssertionError):
+            m.assert_not_called()
+
+    @aioresponses()
+    def test_assert_called(self, m: aioresponses):
+        m.get(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_called()
+        self.run_async(self.session.get(self.url))
+
+        m.assert_called_once()
+        m.assert_called_once_with(self.url)
+        m.assert_called_with(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_not_called()
+
+        with self.assertRaises(AssertionError):
+            m.assert_called_with("http://foo.bar")
+
+    @aioresponses()
+    async def test_assert_called_twice(self, m: aioresponses):
+        m.get(self.url, repeat=True)
+        m.assert_not_called()
+        await self.session.get(self.url)
+        await self.session.get(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_called_once()
+
+    @aioresponses()
+    async def test_integer_repeat_once(self, m: aioresponses):
+        m.get(self.url, repeat=1)
+        m.assert_not_called()
+        await self.session.get(self.url)
+        with self.assertRaises(ClientConnectionError):
+            await self.session.get(self.url)
+
+    @aioresponses()
+    async def test_integer_repeat_twice(self, m: aioresponses):
+        m.get(self.url, repeat=2)
+        m.assert_not_called()
+        await self.session.get(self.url)
+        await self.session.get(self.url)
+        with self.assertRaises(ClientConnectionError):
+            await self.session.get(self.url)
+
+    @aioresponses()
+    async def test_assert_any_call(self, m: aioresponses):
+        http_bin_url = "http://httpbin.org"
+        m.get(self.url)
+        m.get(http_bin_url)
+        await self.session.get(self.url)
+        response = await self.session.get(http_bin_url)
+        self.assertEqual(response.status, 200)
+        m.assert_any_call(self.url)
+        m.assert_any_call(http_bin_url)
+
+    @aioresponses()
+    async def test_assert_any_call_not_called(self, m: aioresponses):
+        http_bin_url = "http://httpbin.org"
+        m.get(self.url)
+        response = await self.session.get(self.url)
+        self.assertEqual(response.status, 200)
+        m.assert_any_call(self.url)
+        with self.assertRaises(AssertionError):
+            m.assert_any_call(http_bin_url)
+
     @aioresponses()
     async def test_exception_requests_are_tracked(self, mocked):
         kwargs = {"json": [42], "allow_redirects": True}
@@ -529,7 +710,7 @@ class AIOResponsesRaiseForStatusSessionTestCase(AsyncTestCase):
         self.assertEqual(cm.exception.message, http.RESPONSES[400][0])
 
     @aioresponses()
-    @skipIf(condition=AIOHTTP_VERSION < '3.4.0',
+    @skipIf(condition=AIOHTTP_VERSION < MINIMUM_AIOHTTP_VERSION,
             reason='aiohttp<3.4.0 does not support raise_for_status '
                    'arguments for requests')
     async def test_do_not_raise_for_status(self, m):
@@ -539,6 +720,21 @@ class AIOResponsesRaiseForStatusSessionTestCase(AsyncTestCase):
 
         self.assertEqual(response.status, 400)
 
+    @aioresponses()
+    @skipIf(condition=AIOHTTP_VERSION < Version('3.9.0'),
+            reason='aiohttp<3.9.0 does not support callable raise_for_status '
+                   'arguments for requests')
+    async def test_callable_raise_for_status(self, m):
+        async def raise_for_status(response: ClientResponse):
+            if response.status >= 400:
+                raise Exception("callable raise_for_status")
+
+        m.get(self.url, status=400)
+        with self.assertRaises(Exception) as cm:
+            await self.session.get(self.url,
+                                   raise_for_status=raise_for_status)
+        self.assertEqual(str(cm.exception), "callable raise_for_status")
+
 
 class AIOResponseRedirectTest(AsyncTestCase):
 
@@ -607,8 +803,6 @@ class AIOResponseRedirectTest(AsyncTestCase):
         self.assertEqual(str(response.url), self.url)
 
     @aioresponses()
-    @skipIf(condition=AIOHTTP_VERSION < '3.1.0',
-            reason='aiohttp<3.1.0 does not add request info on response')
     async def test_request_info(self, rsps):
         rsps.get(self.url, status=200)
 
@@ -619,8 +813,6 @@ class AIOResponseRedirectTest(AsyncTestCase):
         assert request_info.headers == {}
 
     @aioresponses()
-    @skipIf(condition=AIOHTTP_VERSION < '3.1.0',
-            reason='aiohttp<3.1.0 does not add request info on response')
     async def test_request_info_with_original_request_headers(self, rsps):
         headers = {"Authorization": "Bearer access-token"}
         rsps.get(self.url, status=200)
@@ -630,3 +822,36 @@ class AIOResponseRedirectTest(AsyncTestCase):
         request_info = response.request_info
         assert str(request_info.url) == self.url
         assert request_info.headers == headers
+
+    @aioresponses()
+    async def test_relative_url_redirect_followed(self, rsps):
+        base_url = "https://httpbin.org"
+        url = f"{base_url}/foo/bar"
+        rsps.get(
+            url,
+            status=307,
+            headers={"Location": "../baz"},
+        )
+        rsps.get(f"{base_url}/baz")
+
+        response = await self.session.get(url, allow_redirects=True)
+
+        self.assertEqual(response.status, 200)
+        self.assertEqual(str(response.url), f"{base_url}/baz")
+        self.assertEqual(len(response.history), 1)
+        self.assertEqual(str(response.history[0].url), url)
+
+    async def test_pass_through_unmatched_requests(self):
+        matched_url = "https://matched_example.org"
+        unmatched_url = "https://httpbin.org/get"
+        params_unmatched = {'foo': 'bar'}
+
+        with aioresponses(passthrough_unmatched=True) as m:
+            m.post(URL(matched_url), status=200)
+            mocked_response = await self.session.post(URL(matched_url))
+            response = await self.session.get(
+                URL(unmatched_url), params=params_unmatched
+            )
+            self.assertEqual(response.status, 200)
+            self.assertEqual(str(response.url), 'https://httpbin.org/get?foo=bar')
+            self.assertEqual(mocked_response.status, 200)
